/*
 * calc_scop.c
 *
 *  Created on: 5 cze 2018
 *      Author: Maciej
 */

#include "calc_scop.h"
#include "isl_tool.h"
#include "str_func.h"
#include <string.h>

static void calculate_islSchedule(calc_scop *calc);
static void calculate_islScheduleNoTuples(calc_scop *calc);
static void calculate_relations_and_delta(calc_scop *calc);

__isl_give calc_scop *calc_scop_from_pet_scop(isl_ctx *ctx, __isl_keep pet_scop *pet){
	calc_scop *calc;


	calc = calc_scop_alloc(ctx);
	if(!calc)
		goto error;

	calc->schedule = pet_scop_get_schedule(pet);
	calc->domain = isl_schedule_get_domain(calc->schedule);
	calc->may_reads = pet_scop_get_may_reads(pet);
	calc->may_writes = pet_scop_get_may_writes(pet);
	calc->must_writes = pet_scop_get_must_writes(pet);

	calc->schedule_map = isl_schedule_get_map(calc->schedule);
	calc->schedule_map = isl_union_map_intersect_domain(calc->schedule_map, isl_union_set_copy(calc->domain));
	calc->schedule_map = isl_union_map_gist_domain(calc->schedule_map, isl_union_set_copy(calc->domain));

	calculate_relations_and_delta(calc);
	calculate_islSchedule(calc);
	//calculate_islScheduleNoTuples(calc);

	return calc;
error:
	calc_scop_free(calc);
	return NULL;
};
//########################################################################################################################

typedef struct loop_dims_info{
	int loopDimSize;
	int dims_counter;
	char (*dimsNames)[8];
} loop_dims_info;

typedef struct schedule_map_info{
	loop_dims_info loop_dims;

	char thumbprint[512];

	int	*order_items;
	int order_items_counter;

	isl_union_map* schedule_mapper;
	isl_union_map* schedule_map;
} schedule_map_info;

typedef struct schedule_map_list_info{
	int counter;
	int scopeDimSize;
	schedule_map_info* schedule_map_list;
	isl_union_map* schedule_map;
} schedule_map_list_info;

static void schedule_map_info_free(schedule_map_info schedule_info);
static void schedule_map_list_info_free(schedule_map_list_info *schedule_list);
static schedule_map_list_info* split_schedule_map( __isl_keep isl_union_map* schedule_map);
static int schedule_map_info_compare(schedule_map_info *A, schedule_map_info *B);
static isl_stat schedule_map_mapper_build(__isl_take isl_map *map, void *user);
static isl_stat schedule_map_list_build(__isl_take isl_map *map, void *user);



static __isl_give calc_scop *calc_scop_normalize_for_mapper(__isl_keep calc_scop *oryginal, __isl_keep schedule_map_info *schedule_info, int scopeDimSize);


calc_scop_list* calc_scop_normalize(__isl_keep calc_scop *oryginal, isl_bool islScheduleOptimize){
	int i;
	calc_scop_list *calc_list = (calc_scop_list*)malloc(sizeof(calc_scop_list));
	schedule_map_list_info *schedule_list;
	isl_union_map *ScheduleMap;

	calc_list->counter = 0;
	calc_list->calc_scop_list = 0;
	calc_list->idNames = 0;

	if(islScheduleOptimize == isl_bool_true)
		ScheduleMap = oryginal->islScheduleMap;
	else
		ScheduleMap = oryginal->schedule_map;

	schedule_list = split_schedule_map(ScheduleMap);


	calc_list->calc_scop_list = (calc_scop**)malloc((schedule_list->counter)*sizeof(calc_scop *));
	calc_list->counter = schedule_list->counter;

	for(i = 0; i < schedule_list->counter; i++){
		calc_list->calc_scop_list[i] = calc_scop_normalize_for_mapper(oryginal, &(schedule_list->schedule_map_list[i]), schedule_list->scopeDimSize);
	}

	schedule_map_list_info_free(schedule_list);

	return calc_list;
}

__isl_null calc_scop_list *calc_scop_list_free(__isl_take calc_scop_list *list){
	int i;

	if(!list) return 0;

	for(i=0; i<list->counter; i++){
		calc_scop_free(list->calc_scop_list[i]);
	}
	free(list->calc_scop_list);
	free(list->idNames);
	free(list);

	return 0;
}


static __isl_give calc_scop *calc_scop_normalize_for_mapper(__isl_keep calc_scop *oryginal, __isl_keep schedule_map_info *schedule_info, int scopeDimSize){
	calc_scop *calc;

	calc = calc_scop_alloc(oryginal->ctx);
	if(!calc)
		goto error;

	calc->scopeDimSize = scopeDimSize;
	calc->dimCount = schedule_info->loop_dims.loopDimSize;
	calc->idNames = (char(*)[8])malloc(calc->dimCount*sizeof(char[8]));

	{
		int i, j=0;
		for(i=0; i< schedule_info->loop_dims.dims_counter; i++){
			if(schedule_info->loop_dims.dimsNames[i][0]!='\0'){
				sprintf(calc->idNames[j], "%s",schedule_info->loop_dims.dimsNames[i]);
				j += 1;
			}
		}
	}

	calc->orig_domain = isl_union_set_copy(oryginal->domain);

	calc->schedule_mapper = isl_union_map_copy(schedule_info->schedule_mapper);
	calc->schedule_map = isl_union_map_apply_range(isl_union_map_reverse(isl_union_map_copy(schedule_info->schedule_mapper)),isl_union_map_copy(schedule_info->schedule_map));

	calc->domain = isl_union_map_domain(isl_union_map_reverse(isl_union_map_intersect_domain(isl_union_map_copy(calc->schedule_mapper), isl_union_set_copy(oryginal->domain))));

	calc->schedule_map = isl_union_map_intersect_domain(calc->schedule_map, isl_union_set_copy(calc->domain));
	calc->schedule_map = isl_union_map_gist_domain(calc->schedule_map, isl_union_set_copy(calc->domain));

	calc->may_reads = isl_union_map_apply_range(isl_union_map_reverse(isl_union_map_copy(calc->schedule_mapper)),isl_union_map_copy(oryginal->may_reads));
	calc->may_writes = isl_union_map_apply_range(isl_union_map_reverse(isl_union_map_copy(calc->schedule_mapper)),isl_union_map_copy(oryginal->may_writes));
	calc->must_writes = isl_union_map_apply_range(isl_union_map_reverse(isl_union_map_copy(calc->schedule_mapper)),isl_union_map_copy(oryginal->must_writes));

	calculate_relations_and_delta(calc);
	calculate_islSchedule(calc);
	//calculate_islScheduleNoTuples(calc);

		return calc;
	error:
		calc_scop_free(calc);
		return NULL;
}



static void schedule_map_info_free(schedule_map_info schedule_info){
	free(schedule_info.loop_dims.dimsNames);
	free(schedule_info.order_items);
	isl_union_map_free(schedule_info.schedule_mapper);
	isl_union_map_free(schedule_info.schedule_map);
}

static void schedule_map_list_info_free(schedule_map_list_info *schedule_list){
	int i;

	for(i = 0; i < schedule_list->counter; i++){
		schedule_map_info_free(schedule_list->schedule_map_list[i]);
	}

	free(schedule_list->schedule_map_list);
	free(schedule_list);
}

static schedule_map_list_info* split_schedule_map( __isl_keep isl_union_map* schedule_map){
	schedule_map_list_info* schedule_list = (schedule_map_list_info*)malloc(sizeof(schedule_map_list_info));
	int i;

	schedule_list->counter = 0;
	schedule_list->scopeDimSize = 0;
	schedule_list->schedule_map = 0;
	schedule_list->schedule_map_list = 0;

	isl_union_map_foreach_map_in_map(schedule_map, schedule_map_list_build, schedule_list);

	for(i = 0; i < schedule_list->counter; i++){
		int j;
		isl_union_map_foreach_map_in_map(schedule_list->schedule_map_list[i].schedule_map, schedule_map_mapper_build, &(schedule_list->schedule_map_list[i]));

		for(j = 0; j < schedule_list->schedule_map_list[i].loop_dims.dims_counter ; j++){
			if(!(schedule_list->schedule_map_list[i].loop_dims.dimsNames[j][0]=='\0'))
				schedule_list->schedule_map_list[i].loop_dims.loopDimSize += 1;
		}

		schedule_list->scopeDimSize = (schedule_list->scopeDimSize) > (schedule_list->schedule_map_list[i].loop_dims.loopDimSize) ? (schedule_list->scopeDimSize) : (schedule_list->schedule_map_list[i].loop_dims.loopDimSize);
	}

	for(i = 1; i < schedule_list->counter; i++){
		int j;
		for(j = i; j - 1 >= 0; j--){
			if(schedule_map_info_compare(&(schedule_list->schedule_map_list[j-1]), &(schedule_list->schedule_map_list[j])) > 0){
				schedule_map_info tmp = schedule_list->schedule_map_list[j-1];
				schedule_list->schedule_map_list[j-1] = schedule_list->schedule_map_list[j];
				schedule_list->schedule_map_list[j] = tmp;
			}
			else{
				break;
			}
		}
	}

	return schedule_list;
}

static int schedule_map_info_compare(schedule_map_info *A, schedule_map_info *B){
	int order_items_counter = (A->order_items_counter) < (B->order_items_counter) ? (A->order_items_counter) : (B->order_items_counter);
	int i;

	for(i=0; i<order_items_counter; i++){
		if(A->order_items[i] != B->order_items[i])
			return A->order_items[i] - B->order_items[i];
	}

	return B->order_items_counter - A->order_items_counter;
}

static isl_stat schedule_map_mapper_build(__isl_take isl_map *map, void *user){
	schedule_map_info *schedule_info = (schedule_map_info*)user;
	const char* tupleName = isl_map_get_tuple_name(map, isl_dim_in);
	int i;

	for(i = schedule_info->loop_dims.dims_counter-1; i>=0; i--){
		if(schedule_info->loop_dims.dimsNames[i][0]=='\0'){
			map = isl_map_remove_dims(map, isl_dim_out, i, 1);
		}
		else{
			map = isl_map_set_dim_name(map, isl_dim_out, i, schedule_info->loop_dims.dimsNames[i]);
		}
	}

	map = isl_map_set_tuple_name(map , isl_dim_out, tupleName);

	if(!schedule_info->schedule_mapper){
		schedule_info->schedule_mapper = isl_union_map_from_map(isl_map_copy(map));
	}
	schedule_info->schedule_mapper = isl_union_map_coalesce(isl_union_map_union(schedule_info->schedule_mapper, isl_union_map_from_map(map)));

	return isl_stat_ok;
}

static isl_stat schedule_map_list_build(__isl_take isl_map *map, void *user){
	schedule_map_list_info* schedule_list = (schedule_map_list_info*)user;

	isl_map *reverse_map = isl_map_renew(isl_map_reverse(isl_map_renew(isl_map_copy(map))));
	int dim_counter = isl_map_dim(reverse_map, isl_dim_in);
	schedule_map_info* schedule_info = 0;
	int order_dim_counter = 0;
	char thumbprint[512];
	int i;

	thumbprint[0] = '\0';
	for(i = 0; i < dim_counter; i++){
		const char *dimName = isl_map_get_dim_name(reverse_map, isl_dim_in, i);
		if(dimName == 0){
			order_dim_counter += 1;
			char *maxValStr = get_pw_aff_value_to_str(isl_set_dim_max(isl_map_domain(isl_map_copy(reverse_map)), i));
			sprintf(&thumbprint[strlen(thumbprint)], "%s,", maxValStr);
			free(maxValStr);
		}
		else{
			break;
		}
	}

	for(i = 0; i < schedule_list->counter; i++){
		if(strcmp(schedule_list->schedule_map_list[i].thumbprint, thumbprint) == 0){
			schedule_info = &(schedule_list->schedule_map_list[i]);
			break;
		}
	}

	if(schedule_info == 0){
		schedule_list->counter += 1;
		schedule_list->schedule_map_list = realloc(schedule_list->schedule_map_list, schedule_list->counter*sizeof(schedule_map_info));
		schedule_info = &(schedule_list->schedule_map_list[schedule_list->counter - 1]);
		schedule_info->order_items = (int*)malloc(order_dim_counter*sizeof(int));

		sprintf(schedule_info->thumbprint, "%s", thumbprint);
		for(i = 0; i < order_dim_counter; i++){
			schedule_info->order_items[i] = get_pw_aff_value_to_int(isl_set_dim_max(isl_map_domain(isl_map_copy(reverse_map)), i));
		}
		schedule_info->order_items_counter = order_dim_counter;
		schedule_info->schedule_map = 0;
		schedule_info->schedule_mapper = 0;

		schedule_info->loop_dims.loopDimSize = 0;
		schedule_info->loop_dims.dims_counter = dim_counter;
		schedule_info->loop_dims.dimsNames = (char(*)[8]) malloc(dim_counter*sizeof(char[8]));
		for(i = 0; i < dim_counter; i++){
			schedule_info->loop_dims.dimsNames[i][0] = '\0';
		}
	}

	for(i = 0; i < dim_counter; i++){
		const char *dimName = isl_map_get_dim_name(reverse_map, isl_dim_in, i);
		if(dimName != 0){
			sprintf(schedule_info->loop_dims.dimsNames[i], "%s", dimName);
		}
	}

	if(!schedule_info->schedule_map){
		schedule_info->schedule_map = isl_union_map_from_map(isl_map_copy(map));
	}
	schedule_info->schedule_map = isl_union_map_coalesce(isl_union_map_union(schedule_info->schedule_map, isl_union_map_from_map(map)));

	isl_map_free(reverse_map);

	return isl_stat_ok;
}


//########################################################################################################################

__isl_give calc_scop *calc_scop_alloc(isl_ctx *ctx){
	calc_scop *calc;

	calc = isl_alloc_type(ctx, calc_scop);

	calc->ctx = ctx;

	calc->dimCount = 0;
	calc->idNames = NULL;

	calc->orig_domain = NULL;
	calc->domain = NULL;
	calc->schedule = NULL;
	calc->may_reads = NULL;
	calc->may_writes = NULL;
	calc->must_writes = NULL;

	calc->schedule_map = NULL;
	calc->schedule_mapper = NULL;
	calc->before = NULL;
	calc->dep_raw = NULL;
	calc->dep_war = NULL;
	calc->dep_waw = NULL;
	calc->relations = NULL;
	calc->delta = NULL;

	calc->islSchedule = NULL;
	calc->islScheduleMap = NULL;

	calc->islScheduleNoTuples = NULL;
	calc->islScheduleMapNoTuples = NULL;

	return calc;
}

__isl_null calc_scop *calc_scop_free(__isl_take calc_scop *calc){
	if(!calc)
		return NULL;

	free(calc->idNames);
	isl_union_set_free(calc->orig_domain);
	isl_union_set_free(calc->domain);
	isl_schedule_free(calc->schedule);
	isl_union_map_free(calc->may_reads);
	isl_union_map_free(calc->may_writes);
	isl_union_map_free(calc->must_writes);

	isl_union_map_free(calc->schedule_map);
	isl_union_map_free(calc->schedule_mapper);
	isl_union_map_free(calc->before);
	isl_union_map_free(calc->dep_raw);
	isl_union_map_free(calc->dep_war);
	isl_union_map_free(calc->dep_waw);
	isl_union_map_free(calc->relations);
	isl_union_set_free(calc->delta);

	isl_schedule_free(calc->islSchedule);
	isl_union_map_free(calc->islScheduleMap);

	isl_schedule_free(calc->islScheduleNoTuples);
	isl_union_map_free(calc->islScheduleMapNoTuples);

	free(calc);

	return NULL;
}

static void calculate_islSchedule(calc_scop *calc){
		calc->islSchedule = isl_union_set_compute_schedule(isl_union_set_copy(calc->domain), isl_union_map_copy(calc->relations),isl_union_map_copy(calc->relations));
		calc->islScheduleMap = isl_schedule_get_map(calc->islSchedule);
}

static void calculate_islScheduleNoTuples(calc_scop *calc){
		isl_union_set *domain = 0;
		isl_union_map *relations = 0;

		isl_union_set_foreach_set_in_set(calc->domain, clear_tuple_name_in_set, &domain);
		if(domain == 0)
			domain = isl_union_set_copy(calc->domain);
		isl_union_map_foreach_map_in_map(calc->relations, clear_tuple_names_in_map, &relations);
		if(relations == 0)
			relations = isl_union_map_copy(calc->relations);

		calc->islScheduleNoTuples = isl_union_set_compute_schedule(isl_union_set_copy(domain), isl_union_map_copy(relations),isl_union_map_copy(relations));
		calc->islScheduleMapNoTuples = isl_schedule_get_map(calc->islScheduleNoTuples);

		isl_union_set_free(domain);
		isl_union_map_free(relations);
}

static void calculate_relations_and_delta(calc_scop *calc){
	calc->before = isl_union_map_lex_lt_union_map(isl_union_map_copy(calc->schedule_map),isl_union_map_copy(calc->schedule_map));

	calc->dep_raw = isl_union_map_intersect(
					isl_union_map_apply_range(isl_union_map_copy(calc->must_writes),isl_union_map_reverse(isl_union_map_copy(calc->may_reads))),
					isl_union_map_copy(calc->before));
	calc->dep_raw = isl_union_map_coalesce(calc->dep_raw);

	calc->dep_waw = isl_union_map_intersect(
					isl_union_map_apply_range(isl_union_map_copy(calc->must_writes),isl_union_map_reverse(isl_union_map_copy(calc->must_writes))),
					isl_union_map_copy(calc->before));
	calc->dep_waw = isl_union_map_coalesce(calc->dep_waw);

	calc->dep_war = isl_union_map_intersect(
					isl_union_map_apply_range(isl_union_map_copy(calc->may_reads),isl_union_map_reverse(isl_union_map_copy(calc->must_writes))),
					isl_union_map_copy(calc->before));
	calc->dep_war = isl_union_map_coalesce(calc->dep_war);

	calc->relations = isl_union_map_union(isl_union_map_copy(calc->dep_raw), isl_union_map_copy(calc->dep_war));
	calc->relations = isl_union_map_union(calc->relations, isl_union_map_copy(calc->dep_waw));
	calc->relations = isl_union_map_coalesce(calc->relations);

	if(isl_union_map_is_empty(calc->relations)==isl_bool_false)	{
		isl_union_map* relations = 0;
		isl_union_map_foreach_map_in_map(calc->relations, clear_tuple_names_in_map, &relations);
		calc->delta = isl_union_map_deltas(relations);
	}
	else{
		calc->delta = isl_union_set_empty(isl_union_map_get_space(calc->relations));
	}

	calc->delta = isl_union_set_remove_divs(calc->delta);
}




